#include <QApplication>
#include <QVBoxLayout>
#include <QMainWindow>
#include <QTreeWidget>
#include <QWidget>
#include <QStringList>
#include <QTreeWidgetItem>
#include <QString>
#include <QMouseEvent>
#include <QLine>
#include <QPalette>
#include <QResizeEvent>
#include <QPoint>
#include <QDebug>


enum LEGEND_ITEM_TYPE
{
  LEGEND_GROUP                = QTreeWidgetItem::UserType + 0,
  LEGEND_LAYER                = QTreeWidgetItem::UserType + 1,
  LEGEND_PROPERTY_GROUP       = QTreeWidgetItem::UserType + 2,
  LEGEND_PROPERTY_ITEM        = QTreeWidgetItem::UserType + 3,
  LEGEND_SYMBOL_GROUP         = QTreeWidgetItem::UserType + 4,
  LEGEND_SYMBOL_ITEM          = QTreeWidgetItem::UserType + 5,
  LEGEND_VECTOR_SYMBOL_ITEM   = QTreeWidgetItem::UserType + 6,
  LEGEND_LAYER_FILE_GROUP     = QTreeWidgetItem::UserType + 7,
  LEGEND_LAYER_FILE           = QTreeWidgetItem::UserType + 8
} ;

class TreeWidget : public QTreeWidget
{
    Q_OBJECT
    public:
    TreeWidget(QWidget *parent = 0)
      : QTreeWidget(parent), dragging(false), dragged(NULL), dropOn(NULL)
    {
      line = new QWidget(viewport());
      hideLine();
      line->setAutoFillBackground(true);
      QPalette pal = line->palette();
      pal.setColor(line->backgroundRole(), Qt::blue);
      line->setPalette(pal);
      setColumnCount(2);
      setHeaderLabels(QStringList() << "Name" << "Info");
      setRootIsDecorated(true);
      int i = 0;
      for (; i<5; ++i) {
        new QTreeWidgetItem(invisibleRootItem(), QStringList(QString("Item %1").arg(i)) << QString("Layer"));
      }
      QTreeWidgetItem *group1 = new QTreeWidgetItem(invisibleRootItem(), QStringList(QString("Item %1").arg(i++)) << QString("Group"), LEGEND_GROUP);
      for (; i<9; ++i) {
        new QTreeWidgetItem(group1, QStringList(QString("Item %1").arg(i)) << QString("Layer"));
      }
      QTreeWidgetItem *group2 = new QTreeWidgetItem(invisibleRootItem(), QStringList(QString("Item %1").arg(i++)) << QString("Group"), LEGEND_GROUP);
      for (; i<13; ++i) {
        new QTreeWidgetItem(group2, QStringList(QString("Item %1").arg(i)) << QString("Layer"));
      }
      setDragDropMode(NoDragDrop);
    }
    void mousePressEvent(QMouseEvent *e)
    {
      if (e->button() == Qt::LeftButton) {
        lastPress = e->pos();
        QTreeWidgetItem *item = itemAt(e->pos());
        dragged = item;
      }
      QTreeWidget::mousePressEvent(e);
    }

    void mouseMoveEvent(QMouseEvent *e)
    {
      if (dragging) {
        QTreeWidgetItem *item = itemAt(e->pos());
        if (item) {
          dropOn = item;
          int item_top = visualItemRect(item).top();
          int item_bottom = visualItemRect(item).bottom();
          int item_middle = ( item_top + item_bottom + 1 ) / 2;
          int item_left =  visualItemRect(item).left();
          if (dropOn != dragged ) {
            if (e->pos().y() < item_middle )
            {
              showLine(item_top+1, item_left);
              dropBottom = false;
            }
            else
            {
              showLine(item_bottom-2, item_left);
              dropBottom = true;
            }
            // dropBottom doesn't care if dropping onto itself.
          }
          else
          {
            hideLine();
          }
        } else if (e->pos().y() >= 0) {
          dropOn = NULL; // Means after items
          showLine(visualItemRect( lastVisibleItem() ).bottom() + 1, 0);
        } else {
          dropOn = NULL;
          hideLine();
        }
        e->accept();
      } else if (e->buttons() == Qt::LeftButton
      && (e->pos() - lastPress).manhattanLength() >= QApplication::startDragDistance()) {
        line->show();
        dragging = true;
      } else {
        QTreeWidget::mouseMoveEvent(e);
      }
    }

    void hideLine()
    {
      line->setGeometry(0, -100, 1, 1);
    }

    void showLine(int y, int left)
    {
      line->setGeometry(left, y, viewport()->width(), 2);
    }

    void mouseReleaseEvent(QMouseEvent *e)
    {
      if (dragging) {
        hideLine();
        if (dragged && !(dragged == dropOn) && !isDescendantOf( dragged, dropOn ) )
        {
          QTreeWidgetItem *draggedParent = dragged->parent() ?  dragged->parent() : invisibleRootItem();
          clearSelection();
          draggedParent->removeChild(dragged);
          if ( dropOn )
          {
            // We're dropping on something
            QTreeWidgetItem *dropOnParent = dropOn->parent() ?  dropOn->parent() : invisibleRootItem();

            if ( dropBottom )
            {
              // Releasing event in bottom half of object. If group, insert into, else after.
              if (dropOn->type() == LEGEND_GROUP )
              {
                // Drop into
                dropOn->insertChild(0, dragged);
              }
              else
              {
                // Drop after
                dropOnParent->insertChild(dropOnParent->indexOfChild( dropOn ) + 1, dragged);
              }
            }
            else
            {
              // Releasing event in top half of object. Insert before, always
              dropOnParent->insertChild(dropOnParent->indexOfChild( dropOn ), dragged);
            }
          }
          else
          {
            // We're not dropping on something, add last of the top level items.
            invisibleRootItem()->insertChild(invisibleRootItem()->childCount(), dragged);
          }

          setCurrentItem(dragged);
        }
        dropOn = dragged = NULL;
        dragging = false;
        e->accept();
      } else {
        QTreeWidget::mouseReleaseEvent(e);
      }
    }

    QTreeWidgetItem *lastVisibleItem()
    {
      QTreeWidgetItem *current;
      QTreeWidgetItem *next;

      current = topLevelItem( topLevelItemCount() - 1 );
      while ( ( next = itemBelow( current ) ) )
      {
        current = next;
      }
      return current;
    }

    bool isDescendantOf(QTreeWidgetItem *parent, QTreeWidgetItem *child)
    {
      // Search depth first
      if ( parent && child )
      {
        if ( parent == child->parent() )
        {
          return true;
        }
        for( int i = 0; i < parent->childCount(); i++ )
        {
          if ( isDescendantOf( parent->child(i), child ) )
          {
            return true;
          }
        }
      }
      return false;
    }

  private:
    QWidget *line;
    QPoint lastPress;
    bool dragging;
    bool dropBottom;
    QTreeWidgetItem *dragged, *dropOn;
};

#include "main.moc"


int main(int argc, char **argv)
{
  QApplication a(argc, argv);
  TreeWidget w;
  w.show();
  return a.exec();
}

